function [means, assignments] = mpi_kmeans(data,K)
% mpi_kmeans - cluster data columns into K groups
%
% [means, assignments] = mpi_kmeans(data,K)
%
% This function performs K-means clustering on the columns of the 
% data matrix, distributing the work across multiple processors
% using MatlabMPI.

% Author: David Ross, Aug. 2006
% $Id$

error(nargchk(2,2,nargin));

ITERATIONS = 100;
[D,N] = size(data);

MPI_Init; % Initialize MPI.
comm = MPI_COMM_WORLD; % Create communicator.
comm_size = MPI_Comm_size(comm); % Get size and rank.
my_rank = MPI_Comm_rank(comm);
SOURCE = 0;

% Create timers, which we will use to determine how much time is spent
% performing each of the respective tasks.
time_total = clock;
time_bcast = 0;
time_reduce = 0;


% Initialize the means to be randomly-selected data points.
% The SOURCE node will pick the means, then broadcast them
% to all the other nodes.
means = []; 
if (my_rank == SOURCE)
    select = randperm(N);
    means = data(:,select(1:K));
end
timer = clock;
means = MPI_Bcast(SOURCE, 1, comm, means); % sync means across all nodes
time_bcast = time_bcast + etime(clock,timer);

% Each node should only keep around it's chunk of the data
data = data(:,(my_rank+1):comm_size:end);

% Now perform EM iterations
for iter = 1:ITERATIONS
    [sums, counts] = recompute_means(data, means);
    timer = clock;
    [sums, counts] = MPI_Reduce(SOURCE, 1e3+iter, comm, @MPI_SUM, ...
        sums, counts);
    time_reduce = time_reduce + etime(clock,timer);
    
    if my_rank == SOURCE
        % if any clusters have zeros members, set their means to
        % randomly-selected data points
        zero_count = find(counts==0);
        if any(zero_count)
            counts(zero_count) = 1;
            select = round((rand(1,length(zero_count))-eps) * size(data,2))+1;
            sums(:,zero_count) = data(:,select);
        end
    
        % recompute the means
        means = sums ./ repmat(counts, [D 1]);
    end

    % synchronize the means across all workers
    timer = clock;
    means = MPI_Bcast(SOURCE, 1e6+iter, comm, means);
    time_bcast = time_bcast + etime(clock, timer);
end

MPI_Finalize; % Finalize Matlab MPI.
disp('SUCCESS');
elapsed_time = etime(clock, time_total);
fprintf(1,'elapsed time: total=%g (bcast=%g, reduce=%g)\n', ...
    elapsed_time, time_bcast, time_reduce);
% % PROFILING OFF
% profsave(profile('info'),['profile_result' num2str(my_rank)]);
% profile off clear;
if (my_rank ~= MatMPI_Host_rank(comm)), exit; end

% only one node needs to compute this
assignments = get_owner(data, means);



%//////////////////////////////////////////////////////////
function [sums, counts] = recompute_means(data, means)
% Given some data vectors (columns) and means, returns, for each 
% cluster kk, the number of data points assigned to that cluster,
% counts(kk), and the sum of the assigned data points sums(:,kk).
% This is like an EM iteration, except we're returning sums and 
% counts, rather that the restimated means (which are sums/counts).

[D,K] = size(means);
counts = zeros(1,K);
sums = zeros(D,K);

owner = get_owner(data, means);
for kk = 1:K
    ind = find(owner==kk);
    counts(kk) = length(ind);
    sums(:,kk) = sum(data(:,ind),2);
end


%//////////////////////////////////////////////////////////
function [owner] = get_owner(data, means)
% get_owner - compute which mean each data column is closest to

distances = l2_distance(means,data);
[junk, owner] = min(distances,[],1);


%//////////////////////////////////////////////////////////
function d = l2_distance(X,Y)
% d = l2_distance(X)
% Compute pairwise l2 distances between column vectors in X
% Warning -- this implementation can suffer from cancellation error.

if nargin == 1

    N = size(X,2);

    lengths = sum(X.^2,1);
    d = repmat(lengths,[N 1]) + repmat(lengths',[1 N]);
    d = d - 2* X'*X;
    
elseif nargin == 2
    XX = sum(X.^2,1);
    YY = sum(Y.^2,1);
    d = repmat(XX', [1 size(Y,2)]) + repmat(YY, [size(X,2) 1]);
    d = d - 2*X'*Y;
    
else
    error('must have 1 or 2 arguments');
end
